from copy import deepcopy

import neo
import numpy as np
import quantities as pq

from ..tools import annotate_overlapping_spikes, compute_sync_rate


class SpikeTrainGenerator:
    """
    Class for generation of spike trains called by the gen_recordings function.
    The list of parameters is in default_params/recordings_params.yaml (spiketrains field).

    Parameters
    ----------
    params :  dict
        Dictionary with parameters to simulate spiketrains. Default values can be retrieved with
        mr.get_default_recordings_params()['spiketrains']
    spiketrains : list of neo.SpikeTrain
        List of neo.SpikeTrain objects to instantiate a SpikeTrainGenerator with existing data
    verbose : bool
        If True, output is verbose
    """

    def __init__(self, params=None, spiketrains=None, seed=None, verbose=False):
        self._verbose = verbose
        self._has_spiketrains = False
        self.params = {}
        if params is None:
            if self._verbose:
                print("Using default parameters")
        if spiketrains is None:
            self.params = deepcopy(params)
            if seed is None:
                seed = np.random.randint(1000)
            if self._verbose:
                print("Spiketrains seed: ", seed)
            self.params["seed"] = seed
            np.random.seed(self.params["seed"])

            if "t_start" not in self.params.keys():
                params["t_start"] = 0
            self.params["t_start"] = params["t_start"] * pq.s
            if "duration" not in self.params.keys():
                params["duration"] = 10
            self.params["t_stop"] = self.params["t_start"] + params["duration"] * pq.s
            if "min_rate" not in self.params.keys():
                params["min_rate"] = 0.1
            self.params["min_rate"] = params["min_rate"] * pq.Hz
            if "ref_per" not in self.params.keys():
                params["ref_per"] = 2
            self.params["ref_per"] = params["ref_per"] * pq.ms
            if "process" not in self.params.keys():
                params["process"] = "poisson"
            self.params["process"] = params["process"]
            if "gamma_shape" not in self.params.keys() and params["process"] == "gamma":
                params["gamma_shape"] = 2
                self.params["gamma_shape"] = params["gamma_shape"]

            if "rates" in self.params.keys():  # all firing rates are provided
                self.params["rates"] = self.params["rates"] * pq.Hz
                self.n_neurons = len(self.params["rates"])
            else:
                rates = []
                types = []
                if "f_exc" not in self.params.keys():
                    params["f_exc"] = 5
                self.params["f_exc"] = params["f_exc"] * pq.Hz
                if "f_inh" not in self.params.keys():
                    params["f_inh"] = 15
                self.params["f_inh"] = params["f_inh"] * pq.Hz
                if "st_exc" not in self.params.keys():
                    params["st_exc"] = 1
                self.params["st_exc"] = params["st_exc"] * pq.Hz
                if "st_inh" not in self.params.keys():
                    params["st_inh"] = 3
                self.params["st_inh"] = params["st_inh"] * pq.Hz
                if "n_exc" not in self.params.keys():
                    params["n_exc"] = 2
                self.params["n_exc"] = params["n_exc"]
                if "n_inh" not in self.params.keys():
                    params["n_inh"] = 1
                self.params["n_inh"] = params["n_inh"]

                for exc in np.arange(self.params["n_exc"]):
                    rate = self.params["st_exc"] * np.random.randn() + self.params["f_exc"]
                    if rate < self.params["min_rate"]:
                        rate = self.params["min_rate"]
                    rates.append(rate)
                    types.append("e")
                for inh in np.arange(self.params["n_inh"]):
                    rate = self.params["st_inh"] * np.random.randn() + self.params["f_inh"]
                    if rate < self.params["min_rate"]:
                        rate = self.params["min_rate"]
                    rates.append(rate)
                    types.append("i")
                self.params["rates"] = rates
                self.params["types"] = types
                self.n_neurons = len(self.params["rates"])

            self.info = params
            self.spiketrains = False
        else:
            self.spiketrains = spiketrains
            self.info = {}
            self._has_spiketrains = True
            if params is not None:
                self.params = deepcopy(params)

    def set_spiketrain(self, idx, spiketrain):
        """
        Sets spike train idx to new spiketrain.

        Parameters
        ----------
        idx : int
            Index of spike train to set
        spiketrain : neo.SpikeTrain
            New spike train

        """
        self.spiketrains[idx] = spiketrain

    def generate_spikes(self):
        """
        Generate spike trains based on default_params of the SpikeTrainGenerator class.
        self.spiketrains contains the newly generated spike trains
        """
        import elephant.spike_train_generation as stg
        import elephant.statistics as stat
        
        if not self._has_spiketrains:
            self.spiketrains = []
            idx = 0
            for n in np.arange(self.n_neurons):
                rate = self.params["rates"][n]
                if self.params["process"] == "poisson":
                    st = stg.homogeneous_poisson_process(rate, self.params["t_start"], self.params["t_stop"])
                elif self.params["process"] == "gamma":
                    st = stg.homogeneous_gamma_process(
                        self.params["gamma_shape"], rate, self.params["t_start"], self.params["t_stop"]
                    )
                self.spiketrains.append(st)
                self.spiketrains[-1].annotate(fr=rate)
                if "n_exc" in self.params.keys() and "n_inh" in self.params.keys():
                    if idx < self.params["n_exc"]:
                        self.spiketrains[-1].annotate(cell_type="E")
                    else:
                        self.spiketrains[-1].annotate(cell_type="I")
                idx += 1

            # check consistency and remove spikes below refractory period
            for idx, st in enumerate(self.spiketrains):
                isi = stat.isi(st)
                idx_remove = np.where(isi < self.params["ref_per"])[0]
                spikes_to_remove = len(idx_remove)
                unit = st.times.units

                while spikes_to_remove > 0:
                    new_times = np.delete(st.times, idx_remove[0]) * unit
                    st = neo.SpikeTrain(new_times, t_start=self.params["t_start"], t_stop=self.params["t_stop"])
                    isi = stat.isi(st)
                    idx_remove = np.where(isi < self.params["ref_per"])[0]
                    spikes_to_remove = len(idx_remove)

                st.annotations = self.spiketrains[idx].annotations
                self.set_spiketrain(idx, st)
        else:
            print("SpikeTrainGenerator initialized with existing spike trains!")

    def add_synchrony(self, idxs, rate=0.05, time_jitt=1 * pq.ms, verbose=False):
        """
        Adds synchronous spikes between pairs of spike trains at a certain rate.

        Parameters
        ----------
        idxs : list or array
            Spike train indexes to add synchrony to
        rate : float
            Rate of added synchrony spike to spike train idxs[1] for each spike of idxs[0]
        time_jitt : quantity
            Maximum time jittering between added spikes
        verbose : bool
            If True output is verbose

        Returns
        -------
        sync_rate : float
            New synchrony rate
        fr1 : quantity
            Firing rate spike train 1
        fr2 : quantity
            Firing rate spike train 2

        """
        idx1 = idxs[0]
        idx2 = idxs[1]
        st1 = self.spiketrains[idx1]
        st2 = self.spiketrains[idx2]
        times1 = st1.times
        times2 = st2.times
        t_start = st2.t_start
        t_stop = st2.t_stop
        unit = times2.units

        sync_rate = compute_sync_rate(times1, times2, time_jitt)
        if sync_rate < rate:
            added_spikes_t1 = 0
            added_spikes_t2 = 0

            spiketrains = [st1, st2]
            curr_overlaps = np.floor(sync_rate * (len(times1) + len(times2)))
            tot_spikes = len(times1) + len(times2)

            # this assumes that: target_overlaps = curr_overlaps + add_overlaps
            add_overlaps = int(np.round((curr_overlaps - rate * tot_spikes) / (rate - 1)))

            # find non-overlappping spikes
            annotate_overlapping_spikes(spiketrains)
            st1_no_idx = np.where(spiketrains[0].annotations["overlap"] == "NO")[0]
            st2_no_idx = np.where(spiketrains[1].annotations["overlap"] == "NO")[0]

            st1_no = times1[st1_no_idx]
            st2_no = times2[st2_no_idx]

            all_times_no_shuffle = np.concatenate((st1_no, st2_no))
            all_times_no_shuffle = all_times_no_shuffle[np.random.permutation(len(all_times_no_shuffle))] * unit

            for t in all_times_no_shuffle:
                if added_spikes_t1 + added_spikes_t2 <= add_overlaps:
                    # check time difference (since they are NO, they most likely won't violate ref_period)
                    if t in times1:
                        t1_jitt = (
                            time_jitt.rescale(unit).magnitude * np.random.rand(1)
                            + t.rescale(unit).magnitude
                            - (time_jitt.rescale(unit) / 2).magnitude
                        )
                        if t1_jitt < t_stop:
                            times2 = np.concatenate((np.array(times2), np.array(t1_jitt)))
                            times2 = times2 * unit
                            added_spikes_t1 += 1
                    elif t in times2:
                        t2_jitt = (
                            time_jitt.rescale(unit).magnitude * np.random.rand(1)
                            + t.rescale(unit).magnitude
                            - (time_jitt.rescale(unit) / 2).magnitude
                        )
                        if t2_jitt < t_stop:
                            times1 = np.concatenate((np.array(times1), np.array(t2_jitt)))
                            times1 = times1 * unit
                            added_spikes_t2 += 1
                else:
                    break
            times1 = np.sort(times1)
            times2 = np.sort(times2)

            # remove spike trains violating ref period
            ref_violations_idxs1 = np.where(np.diff(times1) < self.params["ref_per"])[0]
            ref_violations_idxs2 = np.where(np.diff(times2) < self.params["ref_per"])[0]

            if len(ref_violations_idxs1) > 0:
                print(f"Remove {len(ref_violations_idxs1)} violations from times1")
                times1 = np.delete(times1, ref_violations_idxs1) * unit
            if len(ref_violations_idxs2) > 0:
                print(f"Remove {len(ref_violations_idxs2)} violations from times2")
                times2 = np.delete(times2, ref_violations_idxs2) * unit

            sync_rate = compute_sync_rate(times1, times2, time_jitt)
            if verbose:
                print(
                    "Added",
                    added_spikes_t1,
                    "spikes to spike train",
                    idxs[0],
                    "and",
                    added_spikes_t2,
                    "spikes to spike train",
                    idxs[1],
                    "Sync rate:",
                    sync_rate,
                )
        else:
            spiketrains = [st1, st2]
            annotate_overlapping_spikes(spiketrains)
            max_overlaps = np.floor(rate * (len(times1) + len(times2)))
            curr_overlaps = np.floor(sync_rate * (len(times1) + len(times2)))
            remove_overlaps = int(curr_overlaps - max_overlaps)
            if curr_overlaps > max_overlaps:
                st1_to_idx = np.where(spiketrains[0].annotations["overlap"] == "TO")[0]
                st2_to_idx = np.where(spiketrains[1].annotations["overlap"] == "TO")[0]
                perm = np.random.permutation(len(st1_to_idx))[:remove_overlaps]
                st1_ovrl_idx = st1_to_idx[perm]
                st2_ovrl_idx = st2_to_idx[perm]
                idx_rm_1 = st1_ovrl_idx[: remove_overlaps // 2]
                idx_rm_2 = st2_ovrl_idx[remove_overlaps // 2 :]
                times1 = np.delete(st1.times, idx_rm_1)
                times1 = times1 * unit
                times2 = np.delete(st2.times, idx_rm_2)
                times2 = times2 * unit
                sync_rate = compute_sync_rate(times1, times2, time_jitt)
                if verbose:
                    print(
                        "Removed",
                        len(idx_rm_1),
                        "spikes from spike train",
                        idxs[0],
                        "and",
                        len(idx_rm_2),
                        "spikes from spike train",
                        idxs[1],
                        "Sync rate:",
                        sync_rate,
                    )

        st1 = neo.SpikeTrain(times1, t_start=t_start, t_stop=t_stop)
        st2 = neo.SpikeTrain(times2, t_start=t_start, t_stop=t_stop)
        st1.annotations = self.spiketrains[idx1].annotations
        st2.annotations = self.spiketrains[idx2].annotations
        self.set_spiketrain(idx1, st1)
        self.set_spiketrain(idx2, st2)

        fr1 = len(st1.times) / st1.t_stop
        fr2 = len(st2.times) / st2.t_stop

        return sync_rate, fr1, fr2



import os
import time
from pathlib import Path

import neo
import numpy as np
import yaml
from packaging.version import parse

from .generators import (RecordingGenerator, SpikeTrainGenerator,
                         TemplateGenerator)
from .tools import get_binary_cat, load_templates, safe_yaml_load


def gen_recordings(
    params=None,
    templates=None,
    tempgen=None,
    spgen=None,
    verbose=True,
    tmp_mode="memmap",
    template_ids=None,
    tmp_folder=None,
    n_jobs=0,
    drift_dicts=None,
):
    """
    Generates recordings.

    Parameters
    ----------
    templates : str
        Path to generated templates
    params : dict or str
        Dictionary containing recording parameters OR path to yaml file containing parameters
    tempgen : TemplateGenerator
        Template generator object
    spgen : SpikeTrainGenerator
        Spike train generator object. If None spike trains are created from params['spiketrains']
    verbose: bool or int
        Determines the level of verbose. If 1 or True, low-level, if 2 high level, if False, not verbose
    tmp_mode : None, 'h5' 'memmap'
        Use temporary file h5 memmap or None
        None is no temporary file
    template_ids: list or None
        If None, templates are selected randomly based on selection rules. If a list of indices is provided, the
        indices are used to select templates (template selection is bypassed)
    tmp_folder: str or Path
        In case of tmp files, you can specify the folder.
        If None, then it is automatic using tempfile.mkdtemp()

    Returns
    -------
    RecordingGenerator
        Generated recording generator object
    """
    t_start = time.perf_counter()
    if isinstance(params, (str, Path)):
        params = Path(params)
        if params.is_file() and params.suffix in [".yaml", ".yml"]:
            params_dict = safe_yaml_load(params)
    elif isinstance(params, dict):
        params_dict = params
    else:
        params_dict = {}

    if "spiketrains" not in params_dict:
        params_dict["spiketrains"] = {}
    if "templates" not in params_dict:
        params_dict["templates"] = {}
    if "recordings" not in params_dict:
        params_dict["recordings"] = {}
    if "cell_types" not in params_dict:
        params_dict["cell_types"] = {}
    if "seeds" not in params_dict:
        params_dict["seeds"] = {}

    if tempgen is None and templates is None:
        raise AttributeError("Provide either 'templates' or 'tempgen' TemplateGenerator object")

    if tempgen is None:
        templates = Path(templates)
        if templates.suffix in [".h5", ".hdf5"]:
            tempgen = load_templates(templates, verbose=verbose)
        else:
            raise AttributeError("'templates' is not an hdf5 file")

    if "spiketrains" in params_dict["seeds"]:
        if params_dict["seeds"]["spiketrains"] is None:
            params_dict["seeds"]["spiketrains"] = np.random.randint(1, 10000)
    else:
        params_dict["seeds"]["spiketrains"] = np.random.randint(1, 10000)

    if "templates" in params_dict["seeds"]:
        if params_dict["seeds"]["templates"] is None:
            params_dict["seeds"]["templates"] = np.random.randint(1, 10000)
    else:
        params_dict["seeds"]["templates"] = np.random.randint(1, 10000)

    if "convolution" in params_dict["seeds"]:
        if params_dict["seeds"]["convolution"] is None:
            params_dict["seeds"]["convolution"] = np.random.randint(1, 10000)
    else:
        params_dict["seeds"]["convolution"] = np.random.randint(1, 10000)

    if "noise" in params_dict["seeds"]:
        if params_dict["seeds"]["noise"] is None:
            params_dict["seeds"]["noise"] = np.random.randint(1, 10000)
    else:
        params_dict["seeds"]["noise"] = np.random.randint(1, 10000)

    if template_ids is not None:
        celltype_params = params_dict["cell_types"]
        celltypes = tempgen.celltypes
        if celltype_params is not None:
            if "excitatory" in celltype_params.keys() and "inhibitory" in celltype_params.keys():
                exc_categories = celltype_params["excitatory"]
                inh_categories = celltype_params["inhibitory"]
                bin_cat = get_binary_cat(celltypes, exc_categories, inh_categories)[template_ids]
                n_exc = int(np.sum(bin_cat == "E"))
                n_inh = int(np.sum(bin_cat == "I"))
            else:
                bin_cat = np.array(["U"] * len(celltypes))
                n_exc = len(celltypes)
                n_inh = 0
        else:
            bin_cat = np.array(["U"] * len(celltypes))
            n_exc = len(celltypes)
            n_inh = 0
        params_dict["spiketrains"]["n_exc"] = n_exc
        params_dict["spiketrains"]["n_inh"] = n_inh
        if verbose:
            print(f"'template_ids' is given. Setting n_exc={n_exc} and n_inh={n_inh}")

    # Generate spike trains
    if spgen is None:
        spgen = SpikeTrainGenerator(
            params_dict["spiketrains"], verbose=verbose, seed=params_dict["seeds"]["spiketrains"]
        )
        spgen.generate_spikes()
    else:
        assert isinstance(spgen, SpikeTrainGenerator), "'spgen' should be a SpikeTrainGenerator object"
        spgen.info["custom"] = True

    params_dict["spiketrains"] = spgen.info
    # Generate recordings
    recgen = RecordingGenerator(spgen, tempgen, params_dict)
    recgen.generate_recordings(
        tmp_mode=tmp_mode,
        tmp_folder=tmp_folder,
        template_ids=template_ids,
        n_jobs=n_jobs,
        verbose=verbose,
        drift_dicts=drift_dicts,
    )

    if verbose >= 1:
        print("Elapsed time: ", time.perf_counter() - t_start)

    return recgen


def gen_spiketrains(params=None, spiketrains=None, seed=None, verbose=False):
    """
    Generates spike trains.

    Parameters
    ----------
    params : str or dict
        Path to parameters yaml file or parameters dictionary
    spiketrains : list
        List of neo.SpikeTrains (alternative to params definition)
    verbose : bool
        If True, the output is verbose

    Returns
    -------
    SpikeTrainGenerator
        Generated spike train generator object
    """
    if params is None:
        assert spiketrains is not None, "Pass either a 'params' or a 'spiketrains' argument"
        assert isinstance(spiketrains, list) and isinstance(
            spiketrains[0], neo.SpikeTrain
        ), "'spiketrains' should be a list of neo.SpikeTrain objects"
        params_dict = {}
    else:
        if isinstance(params, (str, Path)):
            params = Path(params)
            if params.is_file() and params.suffix in [".yaml", ".yml"]:
                params_dict = safe_yaml_load(params)
        elif isinstance(params, dict):
            params_dict = params
        else:
            params_dict = {}
        spiketrains = None

    spgen = SpikeTrainGenerator(params=params_dict, spiketrains=spiketrains, seed=seed, verbose=verbose)
    spgen.generate_spikes()

    return spgen


def gen_templates(
    cell_models_folder,
    params=None,
    templates_tmp_folder=None,
    tempgen=None,
    intraonly=False,
    parallel=True,
    n_jobs=None,
    joblib_backend="loky",
    recompile=False,
    delete_tmp=True,
    verbose=True,
):
    """

    Parameters
    ----------
    cell_models_folder : str
        path to folder containing cell models
    params : str or dict
        Path to parameters yaml file or parameters dictionary
    templates_tmp_folder: str
        Path to temporary folder where templates are temporarily saved
    tempgen :  TemplateGenerator
        If a TemplateGenerator is passed, the cell types, locations, and rotations of the templates will be set using
        the provided templates
    intraonly : bool
        if True, only intracellular simulation is run
    parallel : bool
        if True, multi-threading is used
    n_jobs: int
        Number of jobs to run in parallel (If None all cpus are used)
    joblib_backend: str
        The joblib backend to use when n_jobs > 1 (default 'loky')
    recompile: bool
        If True, cell models are recompiled
    delete_tmp :
        if True, the temporary files are deleted
    verbose : bool, or int
        If True, the output is verbose (1). If verbose is 2, every step produces an output

    Returns
    -------
    TemplateGenerator
        Generated template generator object

    """
    if isinstance(params, (str, Path)):
        params = Path(params)
        if params.is_file() and params.suffix in [".yaml", ".yml"]:
            params_dict = safe_yaml_load(params)
    elif isinstance(params, dict):
        params_dict = params
    else:
        params_dict = None

    if tempgen is not None:
        if isinstance(tempgen, (str, Path)):
            tempgen_arg = load_templates(tempgen)
        else:
            assert isinstance(tempgen, TemplateGenerator), "'tempgen' should be a TemplateGenerator"
            tempgen_arg = tempgen
    else:
        tempgen_arg = tempgen

    if templates_tmp_folder is not None:
        if not Path(templates_tmp_folder).is_dir():
            os.makedirs(templates_tmp_folder)

    tempgen = TemplateGenerator(
        cell_models_folder=cell_models_folder,
        params=params_dict,
        templates_folder=templates_tmp_folder,
        tempgen=tempgen_arg,
        intraonly=intraonly,
        parallel=parallel,
        recompile=recompile,
        n_jobs=n_jobs,
        joblib_backend=joblib_backend,
        delete_tmp=delete_tmp,
        verbose=verbose,
    )
    tempgen.generate_templates()

    return tempgen



"""
This module group several sub function for recording generator.

All this function work on chunk of signals.
They can be call in loop mode or with joblib.

Important:

When tmp_mode=='memmap' : these functions must assign and add directly the buffer.
When tmp_mode is None : these functions return the buffer and the assignament is done externally.



"""
import h5py
import numpy as np
import scipy.signal

from MEArec.tools import (compute_drift_idxs_from_drift_list,
                          convolve_single_template,
                          convolve_templates_spiketrains,
                          filter_analog_signals)


class FuncThenAddChunk:
    """
    Helper function to compute and add to an existing array by chunk.

    E.g., used for convolution, additive noise
    """

    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kargs):
        return_dict = self.func(*args)

        (
            ch,
            i_start,
            i_stop,
        ) = args[:3]

        assignment_dict = kargs["assignment_dict"]
        tmp_mode = kargs["tmp_mode"]

        if tmp_mode is None:
            pass
        elif tmp_mode == "memmap":
            for key, full_arr in assignment_dict.items():
                out_chunk = return_dict.pop(key)
                full_arr[i_start:i_stop] += out_chunk.astype(full_arr.dtype)

        return return_dict


class FuncThenReplaceChunk:
    """
    Helper function to compute by and replace an existing array by chunk.

    E.g., used for convolution, additive noise
    """

    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kargs):
        return_dict = self.func(*args)

        (
            ch,
            i_start,
            i_stop,
        ) = args[:3]

        assignment_dict = kargs["assignment_dict"]
        tmp_mode = kargs["tmp_mode"]

        if tmp_mode is None:
            pass
        elif tmp_mode == "memmap":
            for key, full_arr in assignment_dict.items():
                out_chunk = return_dict.pop(key)
                full_arr[i_start:i_stop] = out_chunk.astype(full_arr.dtype)

        return return_dict


def chunk_convolution_(
    ch,
    i_start,
    i_stop,
    fs,
    lsb,
    st_idxs,
    pad_samples,
    modulation,
    drifting,
    drifting_units,
    templates,
    cut_outs_samples,
    drift_list,
    verbose,
    amp_mod,
    bursting_units,
    shape_mod,
    shape_stretch,
    extract_spike_traces,
    voltage_peaks,
    dtype,
    seed_list,
):
    """
    Perform full convolution for all spike trains by chunk.

    Parameters
    ----------
    ch: int
        Chunk id
    i_start: int
        first index of chunk
    i_stop:
        last index of chunk (exclude)
    fs: float
        Sampling frequency in Hz
    lsb: int or None
        If given, the signal is quantized given the lsb value
    st_idxs: list
        list of spike indexes for each spike train
    pad_samples: int
        Number of samples for padding
    modulation: str
        Modulation type
    drifting: bool
        If True drifting is performed
    drifting_units: list
        List of drifting units (if None all units are drifted)
    templates: np.array
        Templates
    cut_outs_samples: list
        List with number of samples to cut out before and after spike peaks
    drift_list: list of dict
        List of drift dictionaries
    verbose: bool
        If True output is verbose
    amp_mod: np.array
        Array with modulation values
    bursting_units : list
        List of bursting units
    shape_mod: bool
        If True waveforms are modulated in shape
    shape_stretch: float
        Low and high frequency for bursting
    extract_spike_traces: bool
        If True (default), spike traces are extracted
    voltage_peaks: np.array
        Array containing the voltage values at the peak
    """
    if verbose:
        print("Start convolutions for chunk", ch)
    # set seed
    np.random.seed(seed_list[ch])
    traces_dtype = np.float32

    # checking boundaries not needed here because the recordings are created on the fly
    i_start_pad = i_start - pad_samples
    i_stop_pad = i_stop + pad_samples
    n_samples = i_stop_pad - i_start_pad

    template_idxs = []
    if extract_spike_traces:
        spike_traces = np.zeros((n_samples, len(st_idxs)), dtype=traces_dtype)
    if len(templates.shape) == 4:
        n_elec = templates.shape[2]
    elif len(templates.shape) == 5:
        n_elec = templates.shape[3]
        drift_steps = templates.shape[1]
        default_drift_ind = drift_steps // 2
    else:
        raise AttributeError("Wrong 'templates' shape!")

    recordings = np.zeros((n_samples, n_elec), dtype=traces_dtype)

    for st, st_idx in enumerate(st_idxs):
        if extract_spike_traces:
            max_electrode = np.argmax(voltage_peaks[st])

        seed = np.random.randint(10000)
        np.random.seed(seed)

        if modulation in ["electrode", "template"]:
            mod_bool = True
            if bursting_units is not None:
                if st in bursting_units and shape_mod:
                    unit_burst = True
                else:
                    unit_burst = False
            else:
                unit_burst = False
            mod_array = amp_mod[st]
        else:  # modulation 'none'
            mod_bool = False
            mod_array = None
            unit_burst = False

        (spike_in_chunk,) = np.nonzero((st_idx >= i_start_pad) & (st_idx < i_stop_pad))

        if len(spike_in_chunk) > 0:
            st_idx_in_chunk_pad = st_idx[spike_in_chunk] - i_start_pad

            if drifting:
                drift_idxs_in_chunk = np.zeros(len(st_idx_in_chunk_pad), dtype="uint16")

                spike_times_in_chunk = st_idx[spike_in_chunk]
                drift_idxs_in_chunk = compute_drift_idxs_from_drift_list(st, spike_times_in_chunk, drift_list, fs)

                # 4d
                template = templates[st]
                if extract_spike_traces:
                    central_template = templates[st, default_drift_ind, :, max_electrode, :]
            else:
                drift_idxs_in_chunk = None
                # drift_fs = None
                if templates.ndim == 4:
                    drift_vector = None
                    # 3d no drift
                    template = templates[st]
                    if extract_spike_traces:
                        central_template = templates[st, :, max_electrode, :]
                elif templates.ndim == 5:
                    drift_vector = None
                    # 3d no drift
                    template = templates[st, default_drift_ind]
                    if extract_spike_traces:
                        central_template = templates[st, default_drift_ind, :, max_electrode, :]
                else:
                    raise Exception(f"templates.shape no 4 or 5 {templates.shape}")

            recordings = convolve_templates_spiketrains(
                st,
                st_idx_in_chunk_pad,
                template,
                n_samples=n_samples,
                cut_out=cut_outs_samples,
                modulation=mod_bool,
                mod_array=mod_array,
                bursting=unit_burst,
                shape_stretch=shape_stretch,
                verbose=verbose,
                recordings=recordings,
                drift_idxs=drift_idxs_in_chunk,
            )
            np.random.seed(seed)
            if extract_spike_traces:
                spike_traces[:, st] = convolve_single_template(
                    st,
                    st_idx_in_chunk_pad,
                    central_template,
                    n_samples=n_samples,
                    cut_out=cut_outs_samples,
                    modulation=mod_bool,
                    mod_array=mod_array,
                    bursting=unit_burst,
                    shape_stretch=shape_stretch,
                )
        else:
            if verbose:
                print("No spikes found in chunk", ch, "for spike train", st)

    if verbose:
        print("Done all convolutions for chunk", ch)

    return_dict = dict()
    if pad_samples > 0:
        recordings_ret = recordings[pad_samples:-pad_samples]
    else:
        recordings_ret = recordings

    if extract_spike_traces:
        return_dict["spike_traces"] = spike_traces[pad_samples:-pad_samples]

    if lsb is not None:
        recordings_ret = np.floor_divide(recordings_ret, lsb) * lsb

    return_dict["recordings"] = recordings_ret.astype(dtype)
    return_dict["template_idxs"] = template_idxs

    return return_dict


chunk_convolution = FuncThenAddChunk(chunk_convolution_)


def chunk_uncorrelated_noise_(
    ch,
    i_start,
    i_stop,
    fs,
    lsb,
    num_chan,
    noise_level,
    noise_color,
    color_peak,
    color_q,
    color_noise_floor,
    dtype,
    seed_list,
):
    np.random.seed(seed_list[ch])
    length = i_stop - i_start
    additive_noise = noise_level * np.random.randn(length, num_chan)

    if noise_color:
        # iir peak filter
        b_iir, a_iir = scipy.signal.iirpeak(color_peak, Q=color_q, fs=float(fs))
        additive_noise = scipy.signal.filtfilt(b_iir, a_iir, additive_noise, axis=0, padlen=1000)
        additive_noise += (
            color_noise_floor
            * np.std(additive_noise)
            * np.random.randn(additive_noise.shape[0], additive_noise.shape[1])
        )
        additive_noise = additive_noise * (noise_level / np.std(additive_noise))

    return_dict = {}
    if lsb is not None:
        additive_noise = np.floor_divide(additive_noise, lsb) * lsb
    return_dict["additive_noise"] = additive_noise.astype(dtype)

    return return_dict


chunk_uncorrelated_noise = FuncThenAddChunk(chunk_uncorrelated_noise_)


def chunk_distance_correlated_noise_(
    ch,
    i_start,
    i_stop,
    fs,
    lsb,
    noise_level,
    cov_dist,
    n_elec,
    noise_color,
    color_peak,
    color_q,
    color_noise_floor,
    dtype,
    seed_list,
):
    np.random.seed(seed_list[ch])
    length = i_stop - i_start

    additive_noise = noise_level * np.random.multivariate_normal(np.zeros(n_elec), cov_dist, size=length)
    if noise_color:
        # iir peak filter
        b_iir, a_iir = scipy.signal.iirpeak(color_peak, Q=color_q, fs=float(fs))
        additive_noise = scipy.signal.filtfilt(b_iir, a_iir, additive_noise, axis=0, padlen=1000)
        additive_noise = additive_noise + color_noise_floor * np.std(additive_noise) * np.random.multivariate_normal(
            np.zeros(n_elec), cov_dist, size=length
        )
    additive_noise = additive_noise * (noise_level / np.std(additive_noise))

    return_dict = {}
    if lsb is not None:
        additive_noise = np.floor_divide(additive_noise, lsb) * lsb
    return_dict["additive_noise"] = additive_noise.astype(dtype)

    return return_dict


chunk_distance_correlated_noise = FuncThenAddChunk(chunk_distance_correlated_noise_)


def chunk_apply_filter_(ch, i_start, i_stop, fs, lsb, recordings, pad_samples, cutoff, order, mode, dtype):
    n_samples = recordings.shape[0]

    # compute padding idxs
    if pad_samples > 0:
        if i_start > pad_samples:
            i_start_pad = i_start - pad_samples
            pad_start_samples = pad_samples
        else:
            i_start_pad = 0
            pad_start_samples = i_start
        if n_samples - i_stop > pad_samples:
            i_stop_pad = i_stop + pad_samples
            pad_stop_samples = pad_samples
        else:
            i_stop_pad = n_samples
            pad_stop_samples = n_samples - i_stop
    else:
        i_start_pad = i_start
        i_stop_pad = i_stop

    if cutoff.size == 1:
        filtered_chunk = filter_analog_signals(
            recordings[i_start_pad:i_stop_pad], freq=cutoff, fs=fs, filter_type="highpass", mode=mode, order=order
        )
    elif cutoff.size == 2:
        if fs / 2.0 < cutoff[1]:
            filtered_chunk = filter_analog_signals(
                recordings[i_start_pad:i_stop_pad],
                freq=cutoff[0],
                fs=fs,
                filter_type="highpass",
                mode=mode,
                order=order,
            )
        else:
            filtered_chunk = filter_analog_signals(
                recordings[i_start_pad:i_stop_pad], freq=cutoff, fs=fs, filter_type="bandpass", mode=mode, order=order
            )

    filtered_chunk = filtered_chunk
    if pad_samples > 0:
        filtered_chunk = filtered_chunk[pad_start_samples : filtered_chunk.shape[0] - pad_stop_samples]

    return_dict = {}
    if lsb is not None:
        filtered_chunk = np.floor_divide(filtered_chunk, lsb) * lsb
    return_dict["filtered_chunk"] = filtered_chunk.astype(dtype)

    return return_dict


chunk_apply_filter = FuncThenReplaceChunk(chunk_apply_filter_)
